#!/usr/bin/env python3
"""

Central store of Locations,
which expands paths and can hook into the dkey system

::

    locs = JGDVLocator()
    locs.update({"blah": "ex/dir", "bloo": "file:>a/b/c.txt"})

    locs.blah            # {cwd}/ex/dir
    locs['{blah}']       # {cwd}/ex/dir
    locator['{blah}/blee']  # {cwd}/ex/dir/blee

    locator.bloo            # {cwd}/a/b/c.txt
    locator['{bloo}']       # {cwd}/a/b/c.txt
    locator['{bloo}/blee']  # Error

    locator[{blah}/{bloo}'] # {cwd}/ex/dir/a/b/c.txt

JGDVLocator has 3 main access methods::

    JGDVLocator.get    : like dict.get
    JGDVLocator.access : Access the Location object
    JGDVLocator.expand : Expand the location(s) into a path

Shorthands::

    Locator.KEY      # Locator.access
    Locator["{KEY}"] # Locator.expand

"""
# ruff: noqa: ARG002
# mypy: disable-error-code="name-defined"
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import itertools as itz
import logging as logmod
import os
import pathlib as pl
import re
import typing
from copy import deepcopy
from re import Pattern
from collections import defaultdict, deque
from uuid import UUID, uuid1
from weakref import ref

# ##-- end stdlib imports

# ##-- 1st party imports
from jgdv import Mixin, Proto
from jgdv.mixins.path_manip import PathManip_m
from jgdv.structs.chainguard import ChainGuard
from jgdv.structs.dkey import DKey, MultiDKey, NonDKey, SingleDKey, ExpInst_d

from .location import Location
from .errors import DirAbsent, LocationError, LocationExpansionError
from ._interface import Location_p, Locator_p, LocationMeta_e
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never, Any
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Mapping

if TYPE_CHECKING:
    import enum
    from jgdv import Maybe, Stack, Queue, FmtStr, Traceback
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, MutableMapping, Hashable

    from . import _interface as API # noqa: N812
    from jgdv.structs.dkey._util._interface import SourceChain_d

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##--| Vars

##--| Body

class SoftFailMultiDKey(MultiDKey, mark="soft.fail"):

    __slots__ = ()

    def exp_generate_alternatives_h(self, sources:SourceChain_d, opts:dict) -> list:
        """ Expands subkeys, to be merged into the main key"""
        targets = []
        for key in self.keys():
            targets.append([ExpInst_d(value=key, fallback=None)] )
        else:
            if not bool(targets):
                targets.append([
                    ExpInst_d(value=f"{self}", literal=True),
                ])
            return targets

class _LocatorGlobal:
    """ A program global stack of locations.
    Provides the enter/exit store for JGDVLocator objects
    """

    _global_locs : ClassVar[list[JGDVLocator]] = []
    _startup_cwd : ClassVar[pl.Path] = pl.Path.cwd()

    @staticmethod
    def stacklen() -> int:
        return len(_LocatorGlobal._global_locs)

    @staticmethod
    def peek() -> Maybe[JGDVLocator]:
        match _LocatorGlobal._global_locs:
            case []:
                return None
            case [*_, x]:
                return x
            case _:
                return None

    @staticmethod
    def push(locs:JGDVLocator) -> None:
        _LocatorGlobal._global_locs.append(locs)

    @staticmethod
    def pop() -> Maybe[JGDVLocator]:
        match _LocatorGlobal._global_locs:
            case []:
                return None
            case [*xs, x]:
                _LocatorGlobal._global_locs = xs
                return x
            case _:
                return None

    def __get__(self, obj:Any, objtype:Maybe[type]=None) -> Maybe[JGDVLocator]:  # noqa: ANN401
        """ use the descriptor protocol to make a pseudo static-variable
        https://docs.python.org/3/howto/descriptor.html
        """
        return _LocatorGlobal.peek()

class _LocatorUtil_m:

    _data    : dict[DKey|str, Location]

    def update(self, extra:dict|ChainGuard|Location|JGDVLocator, *, strict:bool=True) -> Self:
        """
          Update the registered locations with a dict, chainguard, or other dootlocations obj.

        when strict=True (default), don't allow overwriting existing locations
        """
        match extra: # unwrap to just a dict
            case dict():
                pass
            case Location():
                extra = {extra.key : extra}
            case ChainGuard():
                return self.update(dict(extra), strict=strict)
            case JGDVLocator():
                return self.update(extra._data, strict=strict)
            case _:
                msg = "Tried to update locations with unknown type"
                raise TypeError(msg, extra)

        raw          = dict(self._data.items())
        base_keys    = set(raw.keys())
        new_keys     = set(extra.keys())
        conflicts    = (base_keys & new_keys)
        if strict and bool(conflicts):
            msg = "Strict Location Update conflicts"
            raise LocationError(msg, conflicts)

        for k,v in extra.items():
            try:
                raw[k] = Location(v)
            except KeyError:
                msg = "Couldn't build a Location"
                raise LocationError(msg, k, v) from None

        logging.debug("Registered New Locations: %s", ", ".join(new_keys))
        self._data = raw
        return self

    def metacheck(self, key:str|API.Key_p, *meta:LocationMeta_e) -> bool:
        """ return True if key provided has any of the metadata flags """
        match key:
            case NonDKey():
                return False
            case DKey() if key in self._data:
                data = self._data[key]
                return any(x in data for x in meta)
            case MultiDKey():
                 for k in key:
                     if k not in self._data:
                         continue
                     data = self._data[key]
                     if not any(x in data for x in meta):
                         return False
            case str():
                return self.metacheck(DKey(key, implicit=True), *meta)
        return False

    def registered(self, *values:str|API.Key_p, task:str="doot", strict:bool=True) -> set:
        """ Ensure the values passed in are registered locations,
          error with DirAbsent if they aren't
        """
        missing = {x for x in values if x not in self} # type: ignore

        if strict and bool(missing):
            msg = "Ensured Locations are missing for %s : %s"
            raise DirAbsent(msg, task, missing)

        return missing

    def normalize(self, path:pl.Path|Location, *, symlinks:bool=False) -> pl.Path:
        """
          Expand a path to be absolute, taking into account the set doot root.
          resolves symlinks unless symlinks=True
        """
        match path:
            case Location() if Location.Marks.earlycwd in path: # type: ignore
                the_path = path.path
                return self._normalize(the_path, root=_LocatorGlobal._startup_cwd) # type: ignore
            case Location():
                the_path = path.path
                return self._normalize(the_path, root=self.root) # type: ignore
            case pl.Path():
                return self._normalize(path, root=self.root) # type: ignore
            case _:
                msg = "Bad type to normalize"
                raise TypeError(msg, path)

    def norm(self, path:pl.Path) -> pl.Path:
        return self.normalize(path)

    def pre_expand(self) -> None:
        """
        Called after updating the Locator,
        it pre-expands any registered keys found in registered Locations
        """
        # TODO
        pass

class _LocatorAccess_m:

    _data    : dict[str|API.Key_p, Location]

    def get(self, key:str|API.Key_p, fallback:Maybe[str|pl.Path]=None) -> Maybe[pl.Path]:
        """
        Behavinng like a dict.get,
        uses Locator.access, but coerces to an unexpanded path

        raises a KeyError when fallback is None
        """
        logging.debug("Locator Get: %s", key)
        match fallback:
            case pl.Path() | None:
                pass
            case str() as x:
                fallback =  pl.Path(x)
            case x:
                msg = "Fallback needs to be a path"
                raise TypeError(msg, x)

        match self.access(key):
            case Location() as x:
                return x.path
            case None if fallback is None:
                msg = "Failed to Access"
                raise KeyError(msg, key)
            case None if fallback:
                return fallback
            case None:
                return None

    def access(self, key:str|API.Key_p) -> Maybe[Location]:
        """
          Access the registered Location associated with 'key'
        """
        logging.debug("Locator Access: %s", key)
        match key:
            case str() if key in self: # type: ignore
                return self._data[key]
            case _:
                return None

    def expand(self, key:Location|pl.Path|API.Key_p|str, *, strict:bool=True, norm:bool=True) -> Maybe[pl.Path]:
        """
        Access the locations mentioned in 'key',
        join them together, and normalize it
        """
        logging.debug("Locator Expand: %s", key)
        coerced : DKey = self._coerce_key(key, strict=strict)
        match coerced.expand(self): # type: ignore
            case None if strict:
                msg = "Strict Expansion of Location failed"
                raise KeyError(msg, key)
            case None:
                return None
            case pl.Path() as x if norm:
                return self.normalize(x) # type: ignore
            case pl.Path() as x:
                return x
            case x:
                msg = "Unknown Response When Expanding Location"
                raise TypeError(msg, key, x)

    def _coerce_key(self, key:Location|API.Key_p|str|pl.Path, *, strict:bool=False) -> DKey:
        """ Coerces a key to a MultiDKey for expansion using DKey's expansion mechanism,
        using self as the source
        """
        match key:
            case Location():
                current = key[1,:]
            case DKey():
                current = f"{key:w}"
            case str():
                current = key
            case pl.Path():
                current = str(key)
            case _:
                msg = "Can't perform initial coercion of key"
                raise TypeError(msg, key)

        match strict:
            case False:
                return DKey['soft.fail'](current, ctor=pl.Path)
            case True:
                return DKey(current, ctor=pl.Path) # type: ignore

##--|

@Proto(Locator_p)
@Mixin(_LocatorAccess_m, _LocatorUtil_m, PathManip_m)
class JGDVLocator(Mapping):
    """
      A managing context for storing and converting Locations to Paths.
      key=value pairs in [[locations]] toml blocks are integrated into it.

      It expands relative paths according to cwd(),
      (or the cwd at program start if the Location has the earlycwd flag)

      Can be used as a context manager to expand from a temp different root.
      In which case the current global loc store is at JGDVLocator.Current

      Locations are of the form:
      key = "meta/vars::path/to/dir/or/file.ext"

      simple locations can be accessed as attributes.
      eg: locs.temp

      more complex locations, with expansions, are accessed as items:
      locs['{temp}/somewhere']
      will expand 'temp' (if it is a registered location)
      """
    Marks     : ClassVar[enum.EnumMeta]   = LocationMeta_e
    Current   : ClassVar[_LocatorGlobal]  = _LocatorGlobal()

    _root     : pl.Path
    _data     : dict[DKey|str, Location]
    _loc_ctx  : Maybe[JGDVLocator]

    access    : Callable
    expand    : Callable
    update    : Callable

    def __init__(self, root:pl.Path) -> None:
        self._root    = root.expanduser().resolve()
        self._data    = {}
        self._loc_ctx = None
        match self.Current:
            case None:
                _LocatorGlobal.push(self)
            case JGDVLocator():
                pass

    def __repr__(self) -> str:
        keys = ", ".join(iter(self))
        return f"<JGDVLocator ({_LocatorGlobal.stacklen()}) : {self.root!s} : ({keys})>"

    def __getattr__(self, key:str) -> Location:
        """
          retrieve the raw named location
          eg: locs.temp
          """
        if key.startswith("__") or key.endswith("__"):
            msg = "Location Access Fail"
            raise AttributeError(msg, key)

        match self.access(key):
            case Location() as x:
                return x
            case _:
                raise AttributeError(key)

    def __getitem__(self, val:DKey|pl.Path|Location|str) -> pl.Path:
        """
        Get the expanded path of a key or location

        eg: doot.locs['{data}/somewhere']
        or: doot.locs[pl.Path('data/{other}/somewhere')]
        or  doot.locs[Location("dir::>a/{diff}/path"]

        """
        return self.expand(val, strict=False)

    def __contains__(self, key:object) -> bool:
        """ Test whether a key is a registered location """
        match key:
            case Location():
                return key in self._data.values()
            case str() | pl.Path():
                return key in self._data
            case _:
                return NotImplemented

    def __bool__(self) -> bool:
        return bool(self._data)

    def __eq__(self, other:object) -> bool:
        match other:
            case JGDVLocator() as loc:
                return id(loc) == id(self)
            case _:
                return False

    def __len__(self) -> int:
        return len(self._data)

    def __iter__(self) -> Generator[str|DKey]:
        """ Iterate over the registered location names """
        return iter(self._data.keys()) # type: ignore

    def __call__(self, new_root:Maybe[pl.Path]=None) -> JGDVLocator:
        """ Create a copied locations object, with a different root """
        new_obj = JGDVLocator(new_root or self._root)
        return new_obj.update(self)

    def __enter__(self) -> JGDVLocator:
        """ replaces the global doot.locs with this locations obj,
        and changes the system root to wherever this locations obj uses as root
        """
        _LocatorGlobal.push(self)
        os.chdir(self._root)
        assert(self.Current is not None)
        return self.Current

    def __exit__(self, exc_type:Maybe[type[Exception]], exc_value:Maybe[Exception], exc_traceback:Maybe[Traceback]) -> bool: # type: ignore
        """ returns the global state to its original, """
        _LocatorGlobal.pop()
        assert(self.Current is not None)
        os.chdir(self.Current._root)
        return False

    def clear(self) -> None:
        self._data.clear()

    @property
    def root(self) -> pl.Path:
        """
          the registered root location
        """
        return self._root



#!/usr/bin/env python3
"""

"""

# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import string
import time
import types
import weakref
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
import sh

# ##-- end 3rd party imports

# ##-- 1st party imports
from jgdv import Mixin
from jgdv._abstract.protocols import SpecStruct_p

# ##-- end 1st party imports

from . import _interface as API  # noqa: N812
from . import errors

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, Generic, cast, assert_type, assert_never
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from jgdv import Ident, FmtStr, Rx, RxStr, Func
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from ._interface import Strang_p

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__file__)
##-- end logging

class StrangFormatter(string.Formatter):
    """
      An Expander/Formatter to extend string formatting with options useful for dkey's
      and doot specs/state.

    """

    def format(self, key:str, /, *args:Any, **kwargs:Any) -> str: # noqa: ANN401
        """ format keys as strings """
        match key:
            case str():
                fmt = key
            case pl.Path():
                raise NotImplementedError()
            case _:
                raise TypeError(errors.FormatterExpansionTypeFail, key)

        result = self.vformat(fmt, args, kwargs)
        return result

    def get_value(self, key:int|str, args:Sequence[Any], kwargs:Any) -> str: # noqa: ANN401
        """ lowest level handling of keys being expanded """
        # This handles when the key is something like '1968'
        if isinstance(key, int) and 0 <= key <= len(args):
            return args[key]

        return kwargs.get(key, key)

    def convert_field(self, value:str, conversion:Maybe[str]) -> str:
        # do any conversion on the resulting object
        match conversion:
            case None:
                return value
            case "s" | "p" | "R" | "c" | "t":
                return str(value)
            case "r":
                return repr(value)
            case "a":
                return ascii(value)
            case _:
                raise ValueError(errors.FormatterConversionUnknownSpec.format(spec=conversion))

    def expanded_str(self, value:Strang_p, *, stop:Maybe[int]=None) -> str:
        """ Create a str with generative marks replaced with generated values

        eg: a.b.c.<gen-uuid> -> a.b.c.<UUID:......>
        """
        raise NotImplementedError()

    def format_subval(self, value:Strang_p, val:str, *, no_expansion:bool=False) -> str:
        match val:
            case str():
                return val
            case UUID() if no_expansion:
                return "<uuid>"
            case UUID():
                return f"<uuid:{val}>"
            case _:
                raise TypeError(errors.FormatterUnkownBodyType, val)



#!/usr/bin/env python3
"""

"""
# Import:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from uuid import UUID, uuid1
# ##-- end stdlib imports

# ##-- types
# isort: off
import abc
import collections.abc
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload

if TYPE_CHECKING:
    from jgdv import Maybe
    from typing import Final
    from typing import ClassVar, Any, LiteralString
    from typing import Never, Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Callable, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    type Logger = logmod.Logger

# isort: on
# ##-- end types

from ._interface import LogLevel_e

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

# Vars:
LoggerClass : Final[type] = logmod.getLoggerClass()
# Body:

class JGDVLogger(LoggerClass):
    """ Basic extension of the logger class

    checks the classvar _levels (intEnum) for additional log levels
    which can be accessed as attributes and items.
    eg: logger.trace(...)
    and: logger['trace'](...)


    A Logger can add prefixes to a logged messages.
    eg:
        logger.set_prefixes('[Test]')
        logger.info('this is a test message')
        # Result :  '[Test] this is a test message'

    """
    _prefixes  : list[str|Callable]
    _colour    : Maybe[str]

    @classmethod
    def install(cls) -> None:
        logmod.setLoggerClass(cls)

    def __init__(self, *args:Any, **kwargs:Any) -> None:  # noqa: ANN401
        super().__init__(*args, **kwargs)
        self._prefixes = []
        self._colour   = None

    def __getattr__(self, attr:str) -> Callable:
        try:
            return ftz.partial(self.log, LogLevel_e[attr])
        except KeyError:
            msg = "Invalid Extension Log Level"
            raise AttributeError(msg, attr) from None

    def __getitem__(self, key:str) -> Callable:
        return self.__getattr__(key)

    ##--| public methods
    def set_colour(self, colour:str) -> None:
        self._colour = self._colour or colour

    def set_prefixes(self, *prefixes:str|Callable) -> None:
        """
        Set prefixes for the logger to add to logged messages
        """
        match prefixes:
            case None | ():
                pass
            case [*xs]:
                self._prefixes =  list(xs)

    def prefix(self, prefix:str|Callable) -> Self:
        """ Create a new logger, with a prefix """
        match prefix:
            case str():
                child = self.getChild(prefix)
            case x if callable(x):
                child = self.getChild(prefix.__name__)
            case _:
                raise TypeError(prefix)

        child.set_prefixes(*self._prefixes, prefix)
        return child

    def getChild(self, name:str) -> Self:  # noqa: N802
        """
        Create a child logger, copying the colour of this logger
        """
        child = super().getChild(name)
        child.set_colour(self._colour)
        return child

    def makeRecord(self, *args:Any, **kwargs:Any) -> logmod.LogRecord:  # noqa: ANN401, N802
        """
        A factory method which can be overridden in subclasses to create
        specialized LogRecords.
        args: name, level, fn, lno, msg, args, exc_info,
        kwargs: func=None, extra=None, sinfo=None
        """
        modified : list = list(args)
        msg_total = []
        for pre in self._prefixes:
            match pre:
                case None:
                    pass
                case str():
                    msg_total.append(pre)
                case x if callable(x):
                    msg_total.append(x())
        else:
            match args[4]:
                case str():
                    msg_total.append(args[4])
                case x:
                    msg_total.append("%s")
                    modified[5] = [args[4],  *args[5]]
            modified[4] = "".join(msg_total)

        rv = super().makeRecord(*modified, **kwargs)
        if self._colour and "colour" not in rv.__dict__:
            rv.__dict__["colour"] = self._colour
        return rv



#!/usr/bin/env python3
"""

"""
# mypy: disable-error-code="misc"
# Imports:
from __future__ import annotations

# ##-- stdlib imports
import datetime
import functools as ftz
import importlib
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
from importlib.metadata import EntryPoint
from uuid import UUID, uuid1

# ##-- end stdlib imports

# ##-- 3rd party imports
from pydantic import field_validator, model_validator

# ##-- end 3rd party imports

# ##-- 1st party imports
from .strang import Strang
from . import _interface as API # noqa: N812
from . import errors
from .processor import StrangBasicProcessor
from .formatter import StrangFormatter
# ##-- end 1st party imports

# ##-- types
# isort: off
import abc
import collections.abc
import typing
from typing import TYPE_CHECKING, cast, assert_type, assert_never
from typing import Generic, NewType, Any, Never, Union
# Protocols:
from typing import Protocol, runtime_checkable
# Typing Decorators:
from typing import no_type_check, final, override, overload
from collections.abc import Callable

UnionTypes = types.UnionType | type(Union[int,None])  # noqa: UP007
if TYPE_CHECKING:
    from jgdv.structs.chainguard import ChainGuard
    import enum
    from jgdv import Maybe, Result, MaybeT
    from typing import Final
    from typing import ClassVar, LiteralString
    from typing import Self, Literal
    from typing import TypeGuard
    from collections.abc import Iterable, Iterator, Generator
    from collections.abc import Sequence, Mapping, MutableMapping, Hashable

    from jgdv._abstract.pre_processable import PreProcessResult

    type CheckType    = type | types.UnionType
    type CheckCancel  = Literal[False]
##--|

# isort: on
# ##-- end types

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

ProtoMeta             : Final[type] = type(Protocol)

TooManyTypesToCheck   : Final[str] = "Too many types to check"
SpecialTypeCheckFail  : Final[str] = "Checking Special Types like generics is not supported yet"
##--|

class CodeReference(Strang, no_register=True):
    """ A reference to a class or function.

    can be created from a string (so can be used from toml),
    or from the actual object (from in python)

    Has the form::

        [cls::]module.a.b.c:ClassName

    Can be built with an imported value directly, and a type to check against

    __call__ imports the reference
    """
    __slots__                                 = ("_check", "_value")

    _processor  : ClassVar                    = StrangBasicProcessor()
    _formatter  : ClassVar                    = StrangFormatter()
    _sections   : ClassVar                    = API.Sections_d(*API.CODEREF_DEFAULT_SECS)
    _check      : Maybe[CheckType]

    @classmethod
    def _pre_process_h[T:CodeReference](cls:type[T], input:Any, *args:Any, strict:bool=False, **kwargs:Any) -> MaybeT[bool, *PreProcessResult[T]]:  # noqa: A002, ANN401, ARG003
        inst_data : dict = {}
        post_data : dict = {}
        match input:
            case str():
                full_str = input
            case Callable():
                split_qual = input.__qualname__.split(".")
                val_iden = ":".join([".".join(split_qual[:-1]), split_qual[-1]])
                full_str = f"{input.__module__}:{val_iden}"
                inst_data['value'] = input
        ##--|
        return False, full_str, inst_data, post_data, None

    def __class_getitem__(cls, *args:Any, **kwargs:Any) -> type:  # noqa: ANN401
        match super().__class_getitem__(*args, **kwargs):
            case type() as x:
                return x
            case types.GenericAlias() as alias:
                pass

        match alias.__args__[0]:
            case types.UnionType() as x:
                annotation = str(x)
            case type() as x:
                annotation = x.__name__
            case x:
                raise TypeError(type(x))
        ##--|
        def force_slots(ns:dict) -> None:
            ns['__slots__'] = ()
        newtype = types.new_class(f"{cls.__name__}[{annotation}]", (alias,), exec_body=force_slots)
        return newtype


    def __init__(self, *args:Any, value:Maybe=None, check:Maybe[CheckType|CheckCancel]=None, **kwargs:Any) -> None:  # noqa: ANN401, ARG002
        super().__init__(**kwargs)
        self._value = value
        match check:
            case False:
                self._check = None
            case None:
                self._check = self.expects_type()
            case type() | types.UnionType():
                self._check = check
            case x:
                raise TypeError(type(x))

    def __call__(self, *, check:Maybe[CheckType|CheckCancel]=None, raise_error:bool=False) -> Result[type, ImportError]:
        """ Tries to import and retrieve the reference,
        and casts errors to ImportErrors
        """
        if self._value is not None:
            return self._value
        try:
            return self._do_import(check=check)
        except ImportError as err:
            if raise_error:
                raise
            return err

    def _do_import(self, *, check:Maybe[CheckType|CheckCancel]=None) -> Any:  # noqa: ANN401
        match self._value:
            case None:
                try:
                    mod = importlib.import_module(self.module)
                    curr = getattr(mod, self.value)
                except ModuleNotFoundError as err:
                    err.add_note(f"Origin: {self}")
                    raise
                except AttributeError as err:
                    raise ImportError(errors.CodeRefImportFailed, str(self), self.value, err.args) from None
                else:
                    self._value = curr
            case _:
                curr = self._value
        ##--|
        self._check_imported_type(check)
        return self._value

    def _check_imported_type(self, check:Maybe[CheckType|CheckCancel]=None) -> None:
        marks        : Maybe[type[API.StrangMarkAbstract_e]]
        check_target : Maybe[CheckType]
        is_callable  : bool
        is_type      : bool

        if self._value is None:
            return

        marks        = self.section(0).marks
        assert(marks is not None)
        is_callable  = callable(self._value)
        is_type      = isinstance(self._value, type)
        check_target = self.expects_type(check)

        if marks.fn in self and not is_callable:  # type: ignore[attr-defined]
            raise ImportError(errors.CodeRefImportNotCallable, self._value, self)

        if marks.cls in self and not is_type: # type: ignore[attr-defined]
            raise ImportError(errors.CodeRefImportNotClass, self._value, self)

        if marks.value in self and (is_type or is_callable):
            raise ImportError(errors.CodeRefImportNotValue, self._value, self)

        match check_target, self._value:
            case None, _:
                return
            case x, type() as val if isinstance(x, type|UnionTypes) and x is not None and issubclass(val, x):
                return
            case type() | types.UnionType(), val if isinstance(val, check_target):
                return
            case _:
                raise ImportError(errors.CodeRefImportUnknownFail, self, check_target)

    def to_alias(self, group:str, plugins:dict|ChainGuard) -> str:
        """ TODO Given a nested dict-like, see if this reference can be reduced to an alias """
        base_alias = str(self)
        match [x for x in plugins[group] if x.value == base_alias]:
            case [x, *_]:
                base_alias = x.name

        return base_alias

    def to_uniq(self, *args:str) -> Never:
        raise NotImplementedError(errors.CodeRefUUIDFail)

    def expects_type(self, *args:Maybe[CheckType|CheckCancel]) -> Maybe[CheckType]:
        match self.cls_annotation(), args:
            case _, [False] | False:
                return None
            case [types.UnionType()|type() as check_type], [*xs]: # Merge types to check
                return cast("types.UnionType", Union[check_type, *[x for x in xs if x]])
            case [types.UnionType()|type() as check_type]: # Use annotation
                return check_type
            case [x, *xs], _: # Too many
                raise ImportError(TooManyTypesToCheck, x, xs)
            case _: # Nothing
                return None



#!/usr/bin/env python3
"""

"""

from __future__ import annotations

from jgdv.structs.strang.errors import StrangError

class DKeyError(StrangError):
    pass

class DecorationMismatch(DKeyError):
    pass
